searchNodes=[{"ref":"Dymo.Tag.html","title":"Dymo.Tag","type":"module","doc":"This module provides functionality dedicated to handling tag data. It essentially aims at maintaining singleton labels in a tags table and exposes helper functions to ease their creation."},{"ref":"Dymo.Tag.html#cast/1","title":"Dymo.Tag.cast/1","type":"function","doc":"Casts attributes into a Tag struct."},{"ref":"Dymo.Tag.html#changeset/1","title":"Dymo.Tag.changeset/1","type":"function","doc":"Makes a changeset suited to manipulate the Tag model. Examples iex&gt; &quot;blue&quot; ...&gt; |&gt; changeset() ...&gt; |&gt; Changeset.apply_changes() ...&gt; |&gt; Map.take([:ns, :label]) %{ns: Ns.root_namespace(), label: &quot;blue&quot;} iex&gt; {:paint, &quot;blue&quot;} ...&gt; |&gt; changeset() ...&gt; |&gt; Changeset.apply_changes() ...&gt; |&gt; Map.take([:ns, :label]) %{ns: :paint, label: &quot;blue&quot;} iex&gt; {&quot;car&quot;, &quot;blue&quot;} ...&gt; |&gt; changeset() ...&gt; |&gt; Changeset.apply_changes() ...&gt; |&gt; Map.take([:ns, :label]) %{ns: :car, label: &quot;blue&quot;} iex&gt; %{ns: :car, label: &quot;blue&quot;} ...&gt; |&gt; changeset() ...&gt; |&gt; Changeset.apply_changes() ...&gt; |&gt; Map.take([:ns, :label]) %{ns: :car, label: &quot;blue&quot;} iex&gt; {&quot;non existent&quot;, &quot;blue&quot;} ...&gt; |&gt; changeset() ...&gt; |&gt; Map.take([:valid?]) %{valid?: false}"},{"ref":"Dymo.Tag.html#find_existing/1","title":"Dymo.Tag.find_existing/1","type":"function","doc":"This function gets an existing tag using its label. If the tag doesn&#39;t exit, it is atomically created. It could be described as a &quot;singleton&quot; helper. Examples iex&gt; %{id: id1a} = Tag.find_or_create!(&quot;novel&quot;) ...&gt; [%{id: id1b}, other] = Tag.find_existing([&quot;novel&quot;, &quot;book&quot;]) ...&gt; {id1a, nil} == {id1b, other} true iex&gt; %{id: id1a} = Tag.find_or_create!({:romance, &quot;novel&quot;}) ...&gt; [%{id: id1b}, other] = Tag.find_existing([{:romance, &quot;novel&quot;}, {:scifi, &quot;book&quot;}]) ...&gt; {id1a, nil} == {id1b, other} true"},{"ref":"Dymo.Tag.html#find_or_create!/1","title":"Dymo.Tag.find_or_create!/1","type":"function","doc":"This function gets an existing tag using its label. If the tag doesn&#39;t exit, it is atomically created. It could be described as a &quot;singleton&quot; helper. Examples iex&gt; %{id: id1a} = Tag.find_or_create!(&quot;novel&quot;) ...&gt; [%{id: id2a}, %{id: id3a}] = Tag.find_or_create!([&quot;article&quot;, &quot;book&quot;]) ...&gt; [%{id: id1b}, %{id: id2b}, %{id: id3b}] = Tag.find_or_create!([&quot;novel&quot;, &quot;article&quot;, &quot;book&quot;]) ...&gt; {id1a, id2a, id3a} == {id1b, id2b, id3b} true iex&gt; %{id: id4a} = Tag.find_or_create!({:romance, &quot;novel&quot;}) ...&gt; [%{id: id5a}, %{id: id6a}] = Tag.find_or_create!([{:romance, &quot;article&quot;}, {:scifi, &quot;book&quot;}]) ...&gt; [%{id: id4b}, %{id: id5b}, %{id: id6b}] = Tag.find_or_create!([{:romance, &quot;novel&quot;}, {:romance, &quot;article&quot;}, {:scifi, &quot;book&quot;}]) ...&gt; {id4a, id5a, id6a} == {id4b, id5b, id6b} true"},{"ref":"Dymo.Tag.html#t:attrs/0","title":"Dymo.Tag.attrs/0","type":"type","doc":"Defines attributes for building this model&#39;s changeset"},{"ref":"Dymo.Tag.html#t:label/0","title":"Dymo.Tag.label/0","type":"type","doc":"Defines a tag&#39;s label, namespaced or not."},{"ref":"Dymo.Tag.html#t:label_or_labels/0","title":"Dymo.Tag.label_or_labels/0","type":"type","doc":"Defines a single flat label or a list of labels, namespaced or not."},{"ref":"Dymo.Tag.html#t:namespaced_label/0","title":"Dymo.Tag.namespaced_label/0","type":"type","doc":"Defines a namespaced label."},{"ref":"Dymo.Tag.html#t:t/0","title":"Dymo.Tag.t/0","type":"type","doc":"Defines simple tags identified by a unique label."},{"ref":"Dymo.Tag.Ns.html","title":"Dymo.Tag.Ns","type":"module","doc":"Describes a tag&#39;s namespace."},{"ref":"Dymo.Tag.Ns.html#cast/1","title":"Dymo.Tag.Ns.cast/1","type":"function","doc":"We use cast/1 to transform and normalize external data into internal data, eg when using Ecto.Changeset.cast. Examples Examples iex&gt; nil |&gt; cast() {:ok, root_namespace()} iex&gt; root_namespace() |&gt; cast() {:ok, root_namespace()} iex&gt; &quot;blue&quot; |&gt; cast() {:ok, :blue} iex&gt; :blue |&gt; cast() {:ok, :blue} iex&gt; &quot;non existent atom&quot; |&gt; cast() :error iex&gt; 3 |&gt; cast() :error iex&gt; [] |&gt; cast() :error"},{"ref":"Dymo.Tag.Ns.html#cast!/1","title":"Dymo.Tag.Ns.cast!/1","type":"function","doc":""},{"ref":"Dymo.Tag.Ns.html#dump/1","title":"Dymo.Tag.Ns.dump/1","type":"function","doc":"The dump/1 function converts data into a format that can be stored in the database using the type described by type/0. Examples iex&gt; nil |&gt; dump() {:ok, &quot;root&quot;} iex&gt; root_namespace() |&gt; dump() {:ok, &quot;root&quot;} iex&gt; :blue |&gt; dump() {:ok, &quot;blue&quot;} iex&gt; &quot;blue&quot; |&gt; dump() :error iex&gt; 3 |&gt; dump() :error iex&gt; [] |&gt; dump() :error"},{"ref":"Dymo.Tag.Ns.html#embed_as/1","title":"Dymo.Tag.Ns.embed_as/1","type":"function","doc":"Dictates how the type should be treated inside embeds. By default, the type is sent as itself, without calling dumping to keep the higher level representation. But it can be set to :dump to it is dumped before encoded. Callback implementation for Ecto.Type.embed_as/1."},{"ref":"Dymo.Tag.Ns.html#equal?/2","title":"Dymo.Tag.Ns.equal?/2","type":"function","doc":"Checks if two terms are semantically equal. Callback implementation for Ecto.Type.equal?/2."},{"ref":"Dymo.Tag.Ns.html#load/1","title":"Dymo.Tag.Ns.load/1","type":"function","doc":"We use load/1 to load data from the database into a normalized form, very much like how cast/1 loads data from the outside world. Examples iex&gt; nil |&gt; load() {:ok, root_namespace()} iex&gt; root_namespace() |&gt; load() {:ok, root_namespace()} iex&gt; &quot;blue&quot; |&gt; load() {:ok, :blue} iex&gt; :blue |&gt; load() {:ok, :blue} iex&gt; &quot;non existent atom&quot; |&gt; load() :error iex&gt; 3 |&gt; load() :error iex&gt; [] |&gt; load() :error"},{"ref":"Dymo.Tag.Ns.html#root_namespace/0","title":"Dymo.Tag.Ns.root_namespace/0","type":"function","doc":"This is a configurable parametter allowing to change the atom that will be used for unspecified or nil namespaces. Examples iex&gt; root_namespace() :root"},{"ref":"Dymo.Tag.Ns.html#type/0","title":"Dymo.Tag.Ns.type/0","type":"function","doc":"The type/0 function describe what underlying type should be used for storage. Examples iex&gt; type() :string"},{"ref":"Dymo.Tag.Ns.html#t:castable_t/0","title":"Dymo.Tag.Ns.castable_t/0","type":"type","doc":""},{"ref":"Dymo.Tag.Ns.html#t:t/0","title":"Dymo.Tag.Ns.t/0","type":"type","doc":""},{"ref":"Dymo.Taggable.html","title":"Dymo.Taggable","type":"module","doc":"Allows to easily make models taggable. To use this module, you can simply leverage the use macro inside your model module: use Dymo.Taggable Doing so will require you to have a join table set up between the model and Dymo.Tag as well as a many_to_many relationship on your model (or use the tag() macro to declare the relationship automatically). For more information about their usage and typings, have a look at the Dymo.Tagger behaviour and the Dymo.TaggerImpl implementation."},{"ref":"Dymo.Taggable.html#add_labels/3","title":"Dymo.Taggable.add_labels/3","type":"function","doc":"Adds labels to the given namespace."},{"ref":"Dymo.Taggable.html#all_labels/2","title":"Dymo.Taggable.all_labels/2","type":"function","doc":"Returns all labels associad with the given schema."},{"ref":"Dymo.Taggable.html#labeled_with/3","title":"Dymo.Taggable.labeled_with/3","type":"function","doc":"Returns all objects of given type labeled with given labels."},{"ref":"Dymo.Taggable.html#labels/2","title":"Dymo.Taggable.labels/2","type":"function","doc":"Returns all labels in the given namespace."},{"ref":"Dymo.Taggable.html#remove_labels/3","title":"Dymo.Taggable.remove_labels/3","type":"function","doc":"Remove labels from the given namespace."},{"ref":"Dymo.Taggable.html#set_labels/3","title":"Dymo.Taggable.set_labels/3","type":"function","doc":"Sets all labels with the given namespace, replacing existing ones."},{"ref":"Dymo.Taggable.html#tags/0","title":"Dymo.Taggable.tags/0","type":"macro","doc":"Use this macro in your Ecto schema to add tags field."},{"ref":"Dymo.Taggable.html#t:t/0","title":"Dymo.Taggable.t/0","type":"type","doc":""},{"ref":"Dymo.Taggable.Protocol.html","title":"Dymo.Taggable.Protocol","type":"protocol","doc":""},{"ref":"Dymo.Taggable.Protocol.html#add_labels/3","title":"Dymo.Taggable.Protocol.add_labels/3","type":"function","doc":""},{"ref":"Dymo.Taggable.Protocol.html#labels/2","title":"Dymo.Taggable.Protocol.labels/2","type":"function","doc":""},{"ref":"Dymo.Taggable.Protocol.html#remove_labels/3","title":"Dymo.Taggable.Protocol.remove_labels/3","type":"function","doc":""},{"ref":"Dymo.Taggable.Protocol.html#set_labels/3","title":"Dymo.Taggable.Protocol.set_labels/3","type":"function","doc":""},{"ref":"Dymo.Taggable.Protocol.html#t:t/0","title":"Dymo.Taggable.Protocol.t/0","type":"type","doc":""},{"ref":"Dymo.Tagger.html","title":"Dymo.Tagger","type":"behaviour","doc":"Defines the functions required for a tagger to be compabible with the Dymo.Taggable macro."},{"ref":"Dymo.Tagger.html#__using__/1","title":"Dymo.Tagger.__using__/1","type":"macro","doc":"Use this module to implements alternative Ecto.Tagger By default, all functions are delegated to Ecto.Tagger but can be overriden"},{"ref":"Dymo.Tagger.html#c:add_labels/2","title":"Dymo.Tagger.add_labels/2","type":"callback","doc":"See Dymo.TaggerImpl.add_labels/{2,3}."},{"ref":"Dymo.Tagger.html#c:add_labels/3","title":"Dymo.Tagger.add_labels/3","type":"callback","doc":"See Dymo.TaggerImpl.add_labels/{2,3}."},{"ref":"Dymo.Tagger.html#c:all_labels/2","title":"Dymo.Tagger.all_labels/2","type":"callback","doc":"See Dymo.TaggerImpl.all_labels/3."},{"ref":"Dymo.Tagger.html#c:all_labels/3","title":"Dymo.Tagger.all_labels/3","type":"callback","doc":"See Dymo.TaggerImpl.all_labels/3."},{"ref":"Dymo.Tagger.html#join_key/1","title":"Dymo.Tagger.join_key/1","type":"function","doc":"A helper function that helps computing the field name to use in taggings queries. Examples iex&gt; Tagger.join_key(Dymo.Post) :post_id iex&gt; Tagger.join_key(%Dymo.Post{}) :post_id iex&gt; Tagger.join_key(Person) :person_id"},{"ref":"Dymo.Tagger.html#join_table/1","title":"Dymo.Tagger.join_table/1","type":"function","doc":"A helper function that generates the join table name to be used for a given schema or model. Examples iex&gt; Tagger.join_table(Dymo.Post) &quot;posts_tags&quot; iex&gt; Tagger.join_table(%Dymo.Post{}) &quot;posts_tags&quot; iex&gt; Tagger.join_table(Person) &quot;people_tags&quot;"},{"ref":"Dymo.Tagger.html#c:labeled_with/4","title":"Dymo.Tagger.labeled_with/4","type":"callback","doc":"See Dymo.labeled_with.labels/4."},{"ref":"Dymo.Tagger.html#c:labeled_with/5","title":"Dymo.Tagger.labeled_with/5","type":"callback","doc":"See Dymo.labeled_with.labels/4."},{"ref":"Dymo.Tagger.html#c:labels/3","title":"Dymo.Tagger.labels/3","type":"callback","doc":"See Dymo.TaggerImpl.labels/{3,4}."},{"ref":"Dymo.Tagger.html#c:labels/4","title":"Dymo.Tagger.labels/4","type":"callback","doc":"See Dymo.TaggerImpl.labels/{3,4}."},{"ref":"Dymo.Tagger.html#c:remove_labels/2","title":"Dymo.Tagger.remove_labels/2","type":"callback","doc":"See Dymo.TaggerImpl.remove_labels/3."},{"ref":"Dymo.Tagger.html#c:remove_labels/3","title":"Dymo.Tagger.remove_labels/3","type":"callback","doc":"See Dymo.TaggerImpl.remove_labels/3."},{"ref":"Dymo.Tagger.html#c:set_labels/2","title":"Dymo.Tagger.set_labels/2","type":"callback","doc":"See Dymo.TaggerImpl.set_labels/3."},{"ref":"Dymo.Tagger.html#c:set_labels/3","title":"Dymo.Tagger.set_labels/3","type":"callback","doc":"See Dymo.TaggerImpl.set_labels/3."},{"ref":"Dymo.Tagger.html#singularize/1","title":"Dymo.Tagger.singularize/1","type":"function","doc":""},{"ref":"Dymo.Tagger.html#t:join_key/0","title":"Dymo.Tagger.join_key/0","type":"type","doc":"A join key is an atom."},{"ref":"Dymo.Tagger.html#t:join_table/0","title":"Dymo.Tagger.join_table/0","type":"type","doc":"A join table name is a string."},{"ref":"Dymo.TaggerImpl.html","title":"Dymo.TaggerImpl","type":"module","doc":"This tagger helps with tagging objects using an ecto-backed repo."},{"ref":"Dymo.TaggerImpl.html#add_labels/3","title":"Dymo.TaggerImpl.add_labels/3","type":"function","doc":"Adds labels to a given instance of a model. Examples iex&gt; %Dymo.Post{title: &quot;Hey&quot;} ...&gt; |&gt; Dymo.repo().insert! ...&gt; |&gt; TaggerImpl.set_labels([{:number, &quot;three&quot;}, {:number, &quot;four&quot;}]) ...&gt; |&gt; TaggerImpl.add_labels({:number, &quot;five&quot;}) ...&gt; |&gt; Map.get(:tags) ...&gt; |&gt; Enum.map(&amp; &amp;1.label) ...&gt; |&gt; Enum.sort() ~w(five four three)"},{"ref":"Dymo.TaggerImpl.html#all_labels/3","title":"Dymo.TaggerImpl.all_labels/3","type":"function","doc":"Generates query for retrieving labels associated with a schema. Examples iex&gt; %Dymo.Post{title: &quot;Hey&quot;} ...&gt; |&gt; Dymo.repo().insert! ...&gt; |&gt; TaggerImpl.set_labels([{:number, &quot;eight&quot;}, {:number, &quot;nine&quot;}]) iex&gt; &quot;taggings&quot; ...&gt; |&gt; TaggerImpl.all_labels(:post_id, ns: :number) ...&gt; |&gt; Dymo.repo().all() [&quot;eight&quot;, &quot;nine&quot;]"},{"ref":"Dymo.TaggerImpl.html#labeled_with/5","title":"Dymo.TaggerImpl.labeled_with/5","type":"function","doc":"Queries models that are tagged with the given labels. Examples iex&gt; %{id: id} = %Dymo.Post{title: &quot;Hey&quot;} ...&gt; |&gt; Dymo.repo().insert! ...&gt; |&gt; TaggerImpl.set_labels([{:number, &quot;ten&quot;}, {:number, &quot;eleven&quot;}]) iex&gt; id == Dymo.Post ...&gt; |&gt; TaggerImpl.labeled_with({:number, &quot;ten&quot;}, &quot;taggings&quot;, :post_id) ...&gt; |&gt; Dymo.repo().all() ...&gt; |&gt; hd ...&gt; |&gt; Map.get(:id) true iex&gt; Dymo.Post ...&gt; |&gt; TaggerImpl.labeled_with({:unknown, &quot;nothing&quot;}, &quot;taggings&quot;, :post_id) ...&gt; |&gt; Dymo.repo().all() []"},{"ref":"Dymo.TaggerImpl.html#labels/4","title":"Dymo.TaggerImpl.labels/4","type":"function","doc":"Generates query for retrieving labels associated with a schema&#39;s instance."},{"ref":"Dymo.TaggerImpl.html#remove_labels/3","title":"Dymo.TaggerImpl.remove_labels/3","type":"function","doc":"Removes labels from a given instance of a model. Examples iex&gt; %{tags: tags} = %Dymo.Post{title: &quot;Hey&quot;} ...&gt; |&gt; Dymo.repo().insert! ...&gt; |&gt; TaggerImpl.set_labels([{:number, &quot;six&quot;}, {:number, &quot;seven&quot;}]) ...&gt; |&gt; TaggerImpl.remove_labels({:number, &quot;six&quot;}) iex&gt; Enum.map(tags, &amp; &amp;1.label) [&quot;seven&quot;]"},{"ref":"Dymo.TaggerImpl.html#set_labels/3","title":"Dymo.TaggerImpl.set_labels/3","type":"function","doc":"Sets the labels associated with an instance of a model, for the given namespace. If any other labels are associated to the given model and namespace, they are discarded if they are not part of the list of passed new labels. Examples iex&gt; post = %Dymo.Post{title: &quot;Hey&quot;} ...&gt; |&gt; Dymo.repo().insert! iex&gt; post ...&gt; |&gt; TaggerImpl.set_labels([{:rank, &quot;one&quot;}, {:rank, &quot;two&quot;}]) ...&gt; |&gt; Map.get(:tags) ...&gt; |&gt; Enum.map(&amp; {&amp;1.ns, &amp;1.label}) [{:rank, &quot;one&quot;}, {:rank, &quot;two&quot;}] iex&gt; post ...&gt; |&gt; TaggerImpl.set_labels({:rank, &quot;officer&quot;}) ...&gt; |&gt; Map.get(:tags) ...&gt; |&gt; Enum.map(&amp; {&amp;1.ns, &amp;1.label}) [{:rank, &quot;officer&quot;}]"},{"ref":"Dymo.TaggerImpl.html#t:join_key/0","title":"Dymo.TaggerImpl.join_key/0","type":"type","doc":""},{"ref":"Dymo.TaggerImpl.html#t:join_table/0","title":"Dymo.TaggerImpl.join_table/0","type":"type","doc":""},{"ref":"readme.html","title":"Dymo","type":"extras","doc":"Dymo For all your labelling and tagging needs!™"},{"ref":"readme.html#warning","title":"Dymo - Warning","type":"extras","doc":"Version 1.0.0 is backward-incompatible with other previous versions. Make sure to read the inline documentation to understand what this means, particularily the namespace fields that are required to be added on the Tag schema."},{"ref":"readme.html#motivations","title":"Dymo - Motivations","type":"extras","doc":"When it comes to polymorphism, it&#39;s always hard to find a match-all solution. Each known implementation have tradeoffs: Real polymorphism (taggable_id, taggable_type on the Taggings model) breaks database level foreign keys, and makes requests slower. Parent entity (Having each model have a entity_id refering to some Entity model) and having tags assigned to the Entity model via entity_id is cumbersome when performing complex queries and reverse queries. Multiple nullable foreign keys on the Taggings model (Eg user_id, post_id) is a very good solution but it could become hard to maintain and enforce if many models are taggable. One join table per taggable model (Eg users_tags and posts_tags) make it impossible to consolidate and search using SQL queries only. This package offers an implementation that fits the 3rd and 4th approaches, and it offers shortcuts and mix tasks to make your life easier."},{"ref":"readme.html#installation","title":"Dymo - Installation","type":"extras","doc":"If available in Hex, the package can be installed by adding dymo to your list of dependencies in mix.exs: def deps do [ {:dymo, &quot;~&gt; 0.1.0&quot;} ] end Then, you can install the Tag migration in your application (Note that for umbrella apps, you&#39;ll need to first cd into the app containing your repo migrations): $ mix dymo.install * creating priv/repo/migrations * creating priv/repo/migrations/20180828154957_create_tags.exs Once done, you should start and make a join table for the model(s) you want to be able to label. There is a mix task for this too! $ mix dymo.join_table MyApp.Post * creating priv/repo/migrations * creating priv/repo/migrations/20180828154958_create_posts_tags.exs Once your database gets migrated, a new table posts_tags will be created. You might want to add the following relationship to your MyApp.Post schema: many_to_many :tags, Dymo.Tag, join_through: &quot;posts_tags&quot;, on_replace: :delete, unique: true Alternativelly, you can simply use the `tags()` macro in your schema declaration, as long as you `use Dymo.Taggable` at the top of your module. Note that you can tweak the migrations. For example, you can rename the posts_tags table to whatever you want (eg. taggings) as long as you consistently specify it when using the Tagger macros: use Dymo.Taggable, join_table: &quot;taggings&quot; Once your database gets migrated, a new table posts_tags will be created. If you follow the directives given by the tasks, you should then have a fully labellable Post model. Congratulations!"},{"ref":"readme.html#using-dymo-taggable","title":"Dymo - Using Dymo.Taggable","type":"extras","doc":"When a module uses Dymo.Taggable, many shortcut functions are meta-programmed into it. It becomes easy to achieve labelling-related tasks. All the examples bellow assyme that a Post module calls use Dymo.Taggable. Editing Labels To set the tags on an instance of a post: post |&gt; Taggable.set_labels(~w(ten eleven), ns: :number) |&gt; Taggable.set_labels([{:car, &quot;Fort&quot;}, {:color, &quot;blue&quot;}]) |&gt; Taggable.set_labels(&quot;Heineken&quot;, ns: :beer) Similarily, you can add / remove labels using Post.add_labels and Post.remove_labels. You can also force labelling to only use existing tags (avoid on-the-fly creation) by passing appropriate options. For example: post |&gt; Taggable.set_labels(~w(ten twelve), ns: :number, create_missing: false) |&gt; Taggable.add_labels(&quot;Pierre&quot;, ns: :name, create_missing: false) Querying Labels To get the labels associated with a given post, you have several options. Using the association directly if you defined it: post |&gt; Repo.preload(:tags) |&gt; Map.get(:tags) |&gt; Enum.map(&amp;(&amp;1.label)) Using the helper function: # If the namespace is unspecified, the `:root` namespace is used. post |&gt; Taggable.labels() # Otherwise, only the labels from the given namespace are returned. post |&gt; Taggable.labels(ns: :number) You can also query models that are tagged with specific labels by doing the following: # Match posts tagged with at least *one* of the tags. Post |&gt; Taggable.labelled_with(~w(ten eleven)) # Match posts tagged with at least *all* the specified tags. Post |&gt; Taggable.labelled_with([{:color, &quot;blue&quot;}, {&quot;Heineken&quot;, :beer}], match_all: true)"},{"ref":"readme.html#notes","title":"Dymo - Notes","type":"extras","doc":"Documentation can be generated with ExDoc and published on HexDocs. Once published, the docs can be found at https://hexdocs.pm/dymo."}]